/***********************************************************************
 Name             : lsq.java
 Autor            : Marcel Lattmann
                    Eidgenoessisch Technische Hochschule Zuerich (ETHZ)
 Letzte Aenderung : 26.7.1999

 Einfaches Kleinstquadrat-Applet
 
 Benoetigt folgende Klassen:
   lsqFrame.class         "Hauptprogramm" des Applets
   Matrix.class	          Lineare Algebra mit Matrizen
   ExampleData.class      Beispieldaten erzeugen oder laden
   Plot.class             Kurve plotten
   plotText.class         Text innerhalb der Kurve ausgeben
	
 History:
************************************************************************/

import java.awt.*;
import java.lang.*;
import java.awt.event.*;

class lsqMain {
 public static void main(String argv[]) {
    lsqFrame f = new lsqFrame(600,600,"Kleinstquadrat",100,Color.lightGray);
  }
}

class lsqFrame extends Frame{
  public Panel buttonPanel = new Panel();  // Panel fuer die Tasten

  Label statusLabel;
  Label koordLabel;
  Label ordnungLabel = new Label("Ordnung:  1");

  Button minus = new Button("-");
  Button plus = new Button("+");
  Choice load = new Choice();
  Button random = new Button("Zufllig");
  Button clear = new Button("Lschen");
  Button input = new Button("Input");

  // public StringBuffer dataString = new StringBuffer(); 

  ScrollPane pane;
  Plot plot1 = new Plot();
  public PopupMenu popup = new PopupMenu();

  String newln;
  double zero=0.00001;	// precision for output display

  int ordnung = 1;      // Ordnung des Polynoms

  public int dragIndex = -1;
	
  public lsqFrame(int width, int height, String title, int nrSegs, Color bgColor) {
    // Set up the Frame with the title
    super(title); 
    setSize(width, height);
    
    GridBagConstraints c = new GridBagConstraints();
    GridBagLayout gb = new GridBagLayout();

    // Set line termination character
    newln = System.getProperty("line.separator");

    // setze die Hintergrundfarbe
    setBackground(bgColor);

    // Setze die Anzahl Liniensegmente fuer das Polygon.
    plot1.linesegs=nrSegs;

    setLayout(gb);
    c.insets = new Insets(5,5,5,5);

    pane = new ScrollPane(ScrollPane.SCROLLBARS_AS_NEEDED);  

    // Handler installieren
    plot1.addMouseListener(new mListener());
    plot1.addMouseMotionListener(new mMotionListener());
    plot1.setSize(800,800);
    plot1.repaint();

    c.gridx = 0; c.gridy = 0;
    c.gridwidth  =  GridBagConstraints.REMAINDER;
    c.gridheight = 20;
    c.weightx = 1.0; c.weighty = 1.0;
    c.fill = GridBagConstraints.BOTH;
    c.anchor = GridBagConstraints.NORTH;
    pane.add(plot1);
    add(pane,c);    

    c.gridx = 0; c.gridy = 20;
    c.gridwidth  =  GridBagConstraints.REMAINDER;
    c.gridheight = 1;
    c.weightx = c.weighty = 0.0;
    c.fill = GridBagConstraints.BOTH;
    c.anchor = GridBagConstraints.WEST;
    //    koordLabel = new Label("                                                 " +
    //"                                                                    ");
    koordLabel= new Label("");
    add(koordLabel,c);    

    c.gridx = 0; c.gridy = 21;
    c.gridwidth  =  GridBagConstraints.REMAINDER;
    c.gridheight = 1;
    c.weightx = c.weighty = 0.0;
    c.fill = GridBagConstraints.BOTH;
    c.anchor = GridBagConstraints.WEST;
    statusLabel = new Label("");
  //  label1 = new Label("                                                 " +
   // "                                                                    ");
    add(statusLabel,c);    
    // Konstruiere das buttonPanel
    buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));

    buttonPanel.add(ordnungLabel);

    minus.addActionListener(new minusActionListener());
    buttonPanel.add(minus);

    plus.addActionListener(new plusActionListener());
    buttonPanel.add(plus);

    load.addItem("Lade Punkte");
    load.addItem("Punktmenge 1");
    load.addItem("Punktmenge 2");
    load.addItem("Punktmenge 3");
    load.addItem("Punktmenge 4");
    load.addItemListener(new loadItemListener());
    buttonPanel.add(load);

    random.addActionListener(new randomActionListener());
    buttonPanel.add(random);
    
    clear.addActionListener(new clearActionListener());
    buttonPanel.add(clear);

    input.addActionListener(new inputActionListener());
    buttonPanel.add(input);

    // Fuege das Button-Panel in das Frame ein
    c.gridx = 0; c.gridy = 22;
    c.gridwidth  =  GridBagConstraints.REMAINDER;
    c.gridheight = 1;
    c.weightx = c.weighty = 0.0;
    c.fill = GridBagConstraints.NONE;
    c.anchor = GridBagConstraints.CENTER;
    add(buttonPanel,c);

    addWindowListener(new wAdapter());

    // Setze die Groesse des Applets
    // Dimension d = getPreferredSize();
    // setSize(d);
    setVisible(true);
  }
 
  public void paint() {
    plot1.repaint();
  }

  public int search(double x, double y, double tol) {
    // Gibt den Index des Elements retour, das innerhalb von tol
    // um den Punkt (x,y) liegt. Wenn kein Punkt gefunden, dann gibt es -1 retour
    // tol wird in Pixeln uebergeben und dann umgerechnet.
    int i =0;
    double tolx, toly;
    Dimension d= plot1.getSize();
    tolx = tol*(plot1.xmax-plot1.xmin)/d.width;
    toly = tol*(plot1.ymax-plot1.ymin)/d.height;
 
    while (i<plot1.P.zeilen) {
      if ((x-tolx < plot1.P.element[i][0]) && (x+tolx > plot1.P.element[i][0]) &&
          (y-toly < plot1.P.element[i][1]) && (y+toly > plot1.P.element[i][1]))
	return i;
      i++;
    }
    return -1;
  }

  void compute(Matrix D){
    // Berechne das Polynom mit Hilfe der QR-Zerlegung
    int i, j, k;
    // Als Wert der Funktion compute() wird ein String mit Steigung, x0,
    // Summe der Fehlerquadrate und max. Fehler zurueckgegeben
    String result;

    // Polynom nur dann berechnen, wenn #Punkte > ordnung 
    if (D.zeilen <= ordnung) {
       plot1.M=new Matrix();
       plot1.repaint();
       statusLabel.setText("");
       return;  
    }
 
    // Erzeuge die Modell-Matrix A aus den Punkten
    j = ordnung+1;
    Matrix A = new Matrix(D.zeilen, j, 1);
    for(k=1; k<A.spalten; k++){
      for(i=0; i<A.zeilen; i++){
	A.element[i][k] = Math.pow(D.element[i][0],k);
      }
    }
    
    // Fuehre die QR-Zerlegung von A durch
    Matrix Q = A.Q();
    Matrix R = Matrix.multiply(Q.transpose(), A);
    Matrix b = Matrix.multiply(Q.transpose(), D.sub(0, D.zeilen-1, 1, 1));
    // truncate the non-square portion...
    b = new Matrix(b.sub(0, R.spalten-1, 0, 0));
    R = new Matrix(R.sub(0, R.spalten-1, 0, R.spalten-1));
    Matrix x = Matrix.divide(b,R);

    // x enthaelt nun die Koeffizienten des Ausgleichpolynoms
    plot1.M = x;
    plot1.repaint();   // Punkte und Kurve neu zeichnen

    // Berechne die Steigung, xo, die Summe der Residuenquadrate
    // sowie das maximale Residuum

    b = D.sub(0, D.zeilen-1, 1, 1);
    Matrix Ax = Matrix.multiply(A,x);
    Matrix r = Matrix.subtract(Ax,b);
    double r2sum = r.sumSquares();

    double rmax = r.max();
    double rmin = r.min();
    if (0-rmin > rmax) {
      rmax = 0-rmin;
    }

    if (ordnung == 1)
      result = "x0: " + fixedString(x.element[0][0]) + 
               "   Steigung: " + fixedString(x.element[1][0]) + 
               "   Fehlersumme: " + fixedString(r2sum) + 
               "   Max. Fehler: " + fixedString(rmax);
    else
      result = "x0: " + fixedString(x.element[0][0]) + 
               "   Fehlersumme: " + fixedString(r2sum) + 
               "   Max. Fehler: " + fixedString(rmax);
    statusLabel.setText(result);
  }
  
  String fixedString(String s) {
    // Wandle einen generischen String in einen String fester Laenge
    String sAdd = new String(s + "                                      ");
    return sAdd.substring(0,14);
  }

  String fixedString(double s) {
    // Wandle einen double in einen String fester Laenge
    String sAdd = new String();
    if (Math.abs(s)<zero)
      return new String("0.0000");
    if(s<zero) {
      s = -s;
      sAdd = new String("-");
    }
    Double sD = new Double(s);
    sAdd = sAdd + sD.toString();
    if(sAdd.length()>10){
      sAdd = sAdd.substring(0,10); }
    sAdd = sAdd + "        ";
    return sAdd.substring(0,6);
  }

  String fixedString(int s) {
    // Wandle einen integer in einen String fester Laenge
    Integer sD = new Integer(s);
    String sAdd = new String(sD.toString());
    sAdd = sAdd + "                                      ";
    return sAdd.substring(0,14);
  }

  //***********************************************************************
  // Interne Klassen, um die MausEvents zu handeln
  class mListener implements MouseListener {
    public void mousePressed(MouseEvent e) {
      int index;
      double x,y;

      // Koordinaten bestimmen und in (x,y) umrechnen
      x = e.getX();  y = e.getY();
      Dimension d= plot1.getSize();
      x = (x-plot1.xcenter)*(plot1.xmax-plot1.xmin)/d.width;
      y = (plot1.ycenter-y)*(plot1.ymax-plot1.ymin)/d.height;
      koordLabel.setText("x: " + fixedString(x) + "y: " + fixedString(y)); 

      // Linke Maustaste gedrueckt -> Neuen Punkt einfuegen
      if (!e.isMetaDown() && !e.isAltDown()){
        // Punkt in der Liste suchen
        index=search(x,y,3);

        if (index <0) {   
          // An dieser Position gibt es noch keinen Punkt
          // Neuen Punkt einfuegen
          Matrix newPoint = new Matrix(1,2);
          newPoint.element[0][0] = x;
          newPoint.element[0][1] = y;

          if (plot1.P.zeilen == 0) 
            plot1.P = newPoint;
 	  else
            plot1.P = plot1.P.appendRows(newPoint);
          compute(plot1.P);
          dragIndex = plot1.P.zeilen-1;
        } else
          // Bestehenden Punkt verschieben
          dragIndex=index;
      } else {
        // Mittlere oder rechte Maustaste -> Punkt loeschen
        // Punkt in der Liste suchen
        index=search(x,y,3);
        if (index >=0) {
          plot1.P = plot1.P.deleteRow(index);
          compute(plot1.P);
        }
      }
    }

    public void mouseReleased(MouseEvent e) {
      dragIndex = -1;
    }
    public void mouseClicked(MouseEvent e) {;}
    public void mouseEntered(MouseEvent e) {;}
    public void mouseExited(MouseEvent e) {
      koordLabel.setText(""); 
    }
  }

  //***********************************************************************
  // Interne Klassen, um die Tasten-Events zu handeln

  class minusActionListener implements ActionListener {
    // Wird aufgerufen, wenn der "-"-Button betaetigt wurde
    public void actionPerformed(ActionEvent e) {
      if (ordnung >1) {
        ordnung--;
        ordnungLabel.setText("Ordnung: " + ordnung);
        compute(plot1.P);
      }        
    }    
  }

  class plusActionListener implements ActionListener {
    // Wird aufgerufen, wenn der "+"-Button betaetigt wurde
    public void actionPerformed(ActionEvent e) {
      if (ordnung <20) {
        ordnung++;
        ordnungLabel.setText("Ordnung: " + ordnung);
        compute(plot1.P);
      }
    }    
  }

  class randomActionListener implements ActionListener {
    // Handler fuer den Random-Button
    public void actionPerformed(ActionEvent e) {
      // Random-Daten auf den Bereich 0-10 normieren
      ExampleData ed = new ExampleData(0,0,10,10);
      plot1.P = ed.v;
      compute(plot1.P);
    }
  }

  class clearActionListener implements ActionListener {
    // Handler fuer den Clear-Button
    public void actionPerformed(ActionEvent e) {
      plot1.clear();
      statusLabel.setText("");
    }  
  }

  class inputActionListener implements ActionListener {
    // Handler fuer den Input-Button
    public void actionPerformed(ActionEvent e) {
      Frame f = new Frame();

      String str = new String();

      for (int i=0; i<plot1.P.zeilen; i++) {
        str = str+fixedString(plot1.P.element[i][0]) + ",  " +
                  fixedString(plot1.P.element[i][1]) + newln;
      }
      Input i = new Input(f, str, Color.lightGray);
      i.show();
      str = i.getData();
      plot1.P = new Matrix(str);
      compute(plot1.P);
    }  
  }


  class loadItemListener implements ItemListener {
    // Handler fuer die Auswahl der zu ladenden Punktemenge
    public void itemStateChanged(ItemEvent e) {
      load.select("Lade Punkte");
      if (e.getItem().equals("Punktmenge 1")) {
        ExampleData ed = new ExampleData(1);
        plot1.P = ed.v;      
        compute(plot1.P);
      } else if (e.getItem().equals("Punktmenge 2")) {
        ExampleData ed = new ExampleData(2);
        plot1.P = ed.v;      
        compute(plot1.P);
      } else if (e.getItem().equals("Punktmenge 3")) {
        ExampleData ed = new ExampleData(3);
        plot1.P = ed.v;      
        compute(plot1.P);
      } else if (e.getItem().equals("Punktmenge 4")) {
        ExampleData ed = new ExampleData(4);
        plot1.P = ed.v;      
        compute(plot1.P);
      }
    }
  }

  //***********************************************************************
  // Interne Klasse, um die MouseMotionEvents zu handeln

  class mMotionListener implements MouseMotionListener { 
    public void mouseDragged(MouseEvent e) {
      int index;
      double x,y;

      if (dragIndex != -1) {
        Dimension d= plot1.getSize();
        x = e.getX();  y = e.getY();
        x = (x-plot1.xcenter)*(plot1.xmax-plot1.xmin)/d.width;
        y = (plot1.ycenter-y)*(plot1.ymax-plot1.ymin)/d.height;
        koordLabel.setText("x: " + fixedString(x) + "y: " + fixedString(y)); 
        plot1.P.element[dragIndex][0] = x;
        plot1.P.element[dragIndex][1] = y;
        compute(plot1.P);
      }
    }

    public void mouseMoved(MouseEvent e) {	
      Dimension d = plot1.getSize();
      double x = e.getX();  double y = e.getY();
      x = (x-plot1.xcenter)*(plot1.xmax-plot1.xmin)/d.width;
      y = (plot1.ycenter-y)*(plot1.ymax-plot1.ymin)/d.height;
      koordLabel.setText("x: " + fixedString(x) + "y: " + fixedString(y)); 
    }
  }

  //***********************************************************************
  // Interne Klasse, um die WindowEvents zu handeln

  class wAdapter extends WindowAdapter {
    public void windowClosing(WindowEvent e) {	
      dispose();
      System.exit(0);
    }
  }
}





